@@ -24,81 +24,26 @@ import types
 import time
 import types
 import urllib
+import warnings
 
 import nudge.json
 import nudge.log
 import nudge.arg as args
+from nudge.renderer import Json
 from nudge.validator import ValidationError
-from nudge.renderer import ExceptionRenderer
 from nudge.json import Dictomatic
+from nudge.error import handle_exception, HTTPException, JsonErrorHandler,\
+    DEFAULT_ERROR_CODE, DEFAULT_ERROR_CONTENT_TYPE, DEFAULT_ERROR_CONTENT, responses
 
-_log = logging.getLogger("nudge.servicepublisher")
+_log = logging.getLogger("nudge.publisher")
 
 __all__ = [
-    'responses',
     'Args',
     'Endpoint',
     'WSGIRequest',
     'ServicePublisher',
-    'HTTPException',
 ]
 
-responses = {
-  100: 'Continue',
-  101: 'Switching Protocols',
-
-  200: 'OK',
-  201: 'Created',
-  202: 'Accepted',
-  203: 'Non-Authoritative Information',
-  204: 'No Content',
-  205: 'Reset Content',
-  206: 'Partial Content',
-
-  300: 'Multiple Choices',
-  301: 'Moved Permanently',
-  302: 'Found',
-  303: 'See Other',
-  304: 'Not Modified',
-  305: 'Use Proxy',
-  306: '(Unused)',
-  307: 'Temporary Redirect',
-
-  400: 'Bad Request',
-  401: 'Unauthorized',
-  402: 'Payment Required',
-  403: 'Forbidden',
-  404: 'Not Found',
-  405: 'Method Not Allowed',
-  406: 'Not Acceptable',
-  407: 'Proxy Authentication Required',
-  408: 'Request Timeout',
-  409: 'Conflict',
-  410: 'Gone',
-  411: 'Length Required',
-  412: 'Precondition Failed',
-  413: 'Request Entity Too Large',
-  414: 'Request-URI Too Long',
-  415: 'Unsupported Media Type',
-  416: 'Requested Range Not Satisfiable',
-  417: 'Expectation Failed',
-  418: 'I\'m a teapot',
-
-  500: 'Internal Server Error',
-  501: 'Not Implemented',
-  502: 'Bad Gateway',
-  503: 'Service Unavailable',
-  504: 'Gateway Timeout',
-  505: 'HTTP Version Not Supported',
-}
-
-DEFAULT_ERROR_CODE = 500
-DEFAULT_ERROR_CONTENT_TYPE = "application/json; charset=UTF-8"
-DEFAULT_ERROR_CONTENT = '{"message": "%s", "code": %i}' % (
-    responses[DEFAULT_ERROR_CODE],
-    DEFAULT_ERROR_CODE,
-)
-   
 def lazyprop(fn):
     attr_name = '_lazy_' + fn.__name__
     @property
@@ -115,14 +60,13 @@ class Endpoint(object):
     sequential = []
     named = {}
 
-    def __init__(self, name=None, method=None, uri=None, function=None, 
+    def __init__(self, name=None, method=None, uri=None, uris=None, function=None, 
                  args=None, exceptions=None, renderer=None):
         # Someday support unicode here, for now only bytestrings.
         assert isinstance(name, str)
         assert isinstance(method, str)
-        assert isinstance(uri, str)
-        assert callable(function), "function must be callable, but was %s" %\
-            type(function)
+        assert (not uris and isinstance(uri, str)) or (not uri and isinstance(uris, types.ListType)), "Endpoints must have either a uri or uris, but not both"
+        assert callable(function) or isinstance(function, str), "function must be callable or a string, but was %s" % type(function)
 
         assert not exceptions or isinstance(exceptions, dict), \
             "exceptions must be a dict, but was type %s" % type(exceptions)
@@ -130,7 +74,7 @@ class Endpoint(object):
 
         self.name = name
         self.method = method
-        self.uri = uri
+        self.uris = [uri] if uri else uris
         self.function = function
         if args:
             self.sequential, self.named = args
@@ -141,8 +85,27 @@ class Endpoint(object):
                 "named must be a dict, but was type %s" % type(self.named)
 
         self.exceptions = exceptions
-        self.renderer = renderer
-        self.regex = re.compile(self.method + self.uri)
+        if renderer:
+            if isinstance(renderer, types.TypeType):
+                # support cases where the renderer was specified by the
+                # type and not an instance - todo: Cache renderer
+                self.renderer = renderer()
+                warnings.warn(
+                    "Endpoint %s was passed an uninstantiated renderer" %\
+                    (name))
+            else:
+                self.renderer = renderer
+        else:
+            # Nudge default renderer
+            self.renderer = Json()
+        self.regexs = [re.compile(self.method + uri) for uri in self.uris]
+
+    def match(self, reqline):
+        for regex in self.regexs:
+            match = regex.match(reqline)
+            if match:
+                return match
+
 
     def __call__(self, *args, **kwargs):
         return self.function(*args, **kwargs)
@@ -150,6 +113,30 @@ class Endpoint(object):
 def _write(req, content):
     req._buffer += content
 
+
+class WSGIHeaders(dict):
+
+    def __init__(self, *args, **kwargs):
+        super(WSGIHeaders, self).__init__()
+        self.update(*args, **kwargs)
+
+    def __getitem__(self, key):
+        return super(WSGIHeaders, self).__getitem__(WSGIHeaders.normalize_name(key))
+
+    def __setitem__(self, key, value):
+        return super(WSGIHeaders, self).__setitem__(WSGIHeaders.normalize_name(key), value)
+
+    def get(self, key, default=None):
+        return super(WSGIHeaders, self).get(WSGIHeaders.normalize_name(key), default)
+
+    def set(self, key, value):
+        return super(WSGIHeaders, self).set(WSGIHeaders.normalize_name(key), value)
+    
+    @staticmethod
+    def normalize_name(n):
+        return n.lower().replace('-','_')
+
+
 class WSGIRequest(object):
 
     def __init__(self, req_dict):
@@ -174,7 +161,7 @@ class WSGIRequest(object):
 
     @lazyprop
     def headers(self):
-        _headers = {}
+        _headers = WSGIHeaders()
         for k,v in self.req.iteritems():
             if k.startswith('HTTP_'):
                 _headers[k.replace('HTTP_', '').lower()] = v
@@ -244,36 +231,6 @@ def redirect(uri, headers=None):
         headers
     )
 
-class JsonErrorHandler(object):
-    """ Default nudge error handler.
-        If Nudge catches an HTTPException, we will try to set the code 
-        from that exception. AssertionErrors are considered normal
-        bad request errors, and 400 is returned. Otherwise this returns
-        500 with {code:500 , message:message} where message is the exception
-        message. """
-    code = DEFAULT_ERROR_CODE
-    content_type = DEFAULT_ERROR_CONTENT_TYPE
-    content = DEFAULT_ERROR_CONTENT
-    headers = {}
-    def __call__(self, exp):
-        code = self.code
-        message = exp.message
-        if isinstance(exp, HTTPException):
-            code = exp.status_code
-            if not message:
-                message = responses[code]
-        elif isinstance(exp, AssertionError):
-            code = 400
-            message = exp.message
-            _log.debug("Assertion error: %s", exp.message)
-        else:
-            message = responses[DEFAULT_ERROR_CODE]
-            _log.exception("Exception handled by JsonErrorHandler")
-        # TODO (maybe) add the rest of the exceptions members to the resp
-        content = {'message': message, 'code':code}
-        return code, self.content_type, \
-            nudge.json.json_encode(content), self.headers
-
 class ServicePublisher(object):
 
     def __init__(self, fallbackapp=None, endpoints=None, \
@@ -363,7 +320,7 @@ class ServicePublisher(object):
             reqline = method + urllib.unquote(req.path)
             match = None
             for endpoint in self._endpoints:
-                match = endpoint.regex.match(reqline)
+                match = endpoint.match(reqline)
                 if match:
                     break
 
@@ -382,7 +339,7 @@ class ServicePublisher(object):
 
             # convert all values in req.arguments from lists to scalars,
             # then combine with path args.
-            arguments = dict((k, v[0]) for k, v in req.arguments.iteritems())
+            arguments = dict((k, v[0]) for k, v in req.arguments.iteritems() if isinstance(v, list))
             inargs = dict(match.groupdict(), **arguments)
 
             # compile positional arguments
@@ -402,24 +359,10 @@ class ServicePublisher(object):
 
             # TODO make sure this works with unicode
             _log.debug(_gen_trace_str(endpoint.function, args, kwargs, result))
-
-            if result == None:
-                raise HTTPException(404)
             
-            if endpoint.renderer:
-                if isinstance(endpoint.renderer, types.TypeType):
-                    # support cases where the renderer was specified by the
-                    # type and not an instance - todo: Cache renderer
-                    endpoint.renderer = endpoint.renderer()
-                r = endpoint.renderer(result)
-                content, content_type, code, extra_headers = \
-                    r.content, r.content_type, r.http_status, r.headers
-            else:
-                # Nudge gives back json by default
-                code = 200
-                content_type = DEFAULT_ERROR_CONTENT_TYPE
-                content = nudge.json.json_encode(result)
-                extra_headers = {}
+            r = endpoint.renderer(result)
+            content, content_type, code, extra_headers = \
+                r.content, r.content_type, r.http_status, r.headers
 
         except (Exception), e:
             error_response = None
@@ -460,21 +403,6 @@ class ServicePublisher(object):
 
         return [final_content + "\r\n"]
 
-def handle_exception(exp, exp_handlers):
-    # Check if this endpoint can handle this exception
-    if exp_handlers and exp.__class__ in exp_handlers:
-        exp_handler = exp_handlers[exp.__class__]
-        if callable(exp_handler):
-            # TODO maybe give e the req and start response, maybe add
-            # a finished var to track if e handled everything
-            return exp_handler(exp)
-        else:
-            # Handle 'simple' tuple based exception handler (not callable)
-            return (exp_handler.code, exp_handler.content_type,
-                    exp_handler.content, exp_handler.headers)
-    _log.exception("Unhandled exception class: %s", exp.__class__)
-    raise exp
-
 def _finish_request(req, start_response, code, content_type, content, headers):
     try:
         if isinstance(content, unicode):
@@ -510,12 +438,6 @@ def _gen_trace_str(f, args, kwargs, res):
     else:
         return "%s(%s%s): %s" % (f.__name__, argsstr, kwargsstr, res)
 
-class HTTPException(Exception):
-
-    def __init__(self, status_code, message=None):
-        self.status_code = status_code
-        self.message = message
-
 def _generate_headers(version, status_code, content_length, headers={}):
     """generates status line + headers"""
     lines = [version + " " + str(status_code) + " " +
@@ -526,7 +448,6 @@ def _generate_headers(version, status_code, content_length, headers={}):
 
     return "\r\n".join(lines) + "\r\n\r\n"
 
-
 def serve(service_description, args=None):
     if not args:
         args = sys.argv
